function [varargout]=IM01(Fname)
% IM01 perform non-linear Isometric Method inversion
%
% demonstration of isometric non-linear inversion method IM
% method design: J.Malek
%        M-code: P.Kolar
%
% created        01   15/05/07
% starting ver.  00   29/03/07
%
% call:  [SOLUTION,CONVERGcRVS]=IMxx(NAME)
%          - read parameters from file NAME.par
%          - read data form file NAME.dat
%          - creates protocol file NAME.rNN (included results)
%          - if required, solution and corvegetion curves are returned
%
%    file NAME.m with forward function is required !!
%    syntax:  DAT=name(PAR) - where PARameters and DATa are column vectors
%
% some messages are displaied during the run
% only "Control Constants" in a frame below are allowed to be changed
%

global prt_out
global dat1
global nFwd iLoop
global parBest Mbest Mstart
global m
global hFwd


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%    Control Constants
%
% Only the values defined in this frame are allowed to be changed
% This values control stopping conditions of IM and should be
% tuned according to the optimum performance of the particular IM inversion.
%
% Keep in mind that changes outside this frame may break down the IM
% algorithm!!

iLoop_max = 5000; % max. number of total IM loops
iTry=20;          % max. number of single random Monte Carlo search
cStop=0.000005;   % stop condition: if Mx < cStop*sqrt(n-m), break, end
timMax=600;       % max. run time (in [s]).

%%%%%%%%%%%%%%%%%%%%%end_of_Control_Constants_definition%%%%%%%%%%%%%%%%%%


nFwd=0;
tic;
iLoop=0;
endMsg='O.K.';
kRed=0;             % step of space reduction

for i=1:nargout
    varargout{i}=NaN;
end


%
% protocol file
%
if nargin ~= 1
    disp('No input name given!');
    return
end

prt_out=open_protocol(Fname);
if nargin ~= 1
    run_prt('wrn','No input name given!');
    run_prt('wrn','Default name (im) will be used.');
end
run_prt('msg',['Name: ',Fname]);


%
% set FWD function
%
cur_dir=pwd;
tName=sprintf('%s.m',Fname);
if ~exist([cur_dir,'\',tName],'file')
    run_prt('err',['Required forward function >>',tName,'<< not exists !!']);
    return
end
eval(['hFwd=@',Fname,';']);


%
% read data
%
[nDat,dat1]=read_data(Fname);


%
% read parameters
%
[m,ParDef]=read_param(Fname);


%
%data initialisation
%
datBest=ones(1,nDat)*Inf;
parBest=ones(1,m)*Inf;
Mbest=Inf;
sv=[];   % step vector
svN=[];  % Norm ofstep vectors
r=1.;    % scaling constant
newSimplex=0;


[par,dat]=make_init(m+1,kRed,ParDef);
% ParDef.start value is defined
if ~all(isnan(ParDef.start))
    par(~isnan(ParDef.start),1)=ParDef.start(~isnan(ParDef.start));
    for i=1:size(par,2);
        dat(:,i)=hFwd(par(:,i));
        nFwd=nFwd+1;
    end

end

%variables for plots
ccN  =[];  % convergention curve  x-coord.
ccM  =[];  % convergention curve  y-coord.
ccPar=[];   % so far the best solution (vector)
ccK  =[];    % order of parameter space reduction

%
% inversion
%

%main loop
while(1)
    iLoop=iLoop+1;
    if iLoop >= iLoop_max % check max Loops
        run_prt('wrnEnd',['iLoop over  (max: ',num2str(iLoop_max),') !']);
        endMsg='WARN';
        break;
    end
    if toc > timMax  % check max elapsed time
        run_prt('wrnEnd',[' Time over (max: ',num2str(timMax),') !']);
        endMsg='WARN';
        break;
    end
    %{
    if ~(mod(iLoop,10))  % message every 10th loop
    str=sprintf('iLoop=% 4d,  iFwd=% 5d,     diff=%e ',iLoop,nFwd,Mbest);
    disp(str);
    end
    %}

    %sorting  solutions
    Mdat=getMisfit(dat);
    [Mdat,order]=sort(Mdat);
    if ~all(order == 1:m+1) % is ordering necessary ?
        parNew(:,order)=par(:,:);
        par=parNew;
        datNew(:,order)=dat(:,:);
        dat=datNew;
    end


    %saving so far the best solution
    if Mbest > Mdat(1);
        Mbest=Mdat(1);
        datBest=dat(:,1);
        parBest=par(:,1);
        str=sprintf('    Find new: iLoop=% 4d,  iFwd=% 5d,     Mbest=%e ',iLoop,nFwd,Mbest);
        disp(str);
        ccN  =[ccN nFwd];
        ccM  =[ccM Mbest];
        ccPar=[ccPar parBest];
        ccK=[ccK kRed];
        if newSimplex
            kRed=max(kRed-1,0);
            newSimplex=0;
        end
    else
        if newSimplex
            kRed=kRed+1;
            newSimplex=0;
        end
    end
    if iLoop==1,  %remeber starting values
        Mstart=Mdat(1);
        ccN  =[ccN nFwd];
        ccM  =[ccM Mbest];
        ccPar=[ccPar parBest];
        ccK=[ccK kRed];
    end


    % eq.(10)
    dk=dat(:,2:m+1);
    dL=dat(:,1);
    dLX=repmat(dL,1,m);
    ek=dk-dLX;

    % solution of eq.(11)
    rank1=rank(ek);
    if rank1 <= 2
        run_prt('wrn',['Rank is ',num2str(rank1),'!']);
        ci = pinv(ek)*(dat1-dL);
    else
        ci=ek\(dat1-dL);
    end

    % eq. (12)
    pL=par(:,1);
    pLX=repmat(pL,1,m);
    pk=par(:,2:m+1);
    qk=pk-pLX;

    % eq. (13b)
    ciX=repmat(ci',m,1);
    px=pL+(r* sum(ciX.*qk,2));

    % check whether is in <parMin; parMax>
    if any(px <= ParDef.min), run_prt('wrn',['min for i=',num2str((px <= ParDef.min)')]);
        px=max(px,ParDef.min);
    end
    if any(px >= ParDef.max), run_prt('wrn',['max for i=',num2str((px >= ParDef.max)')]);
        px=min(px,ParDef.max);
    end

    dx=hFwd(px);
    nFwd=nFwd+1;

    Mx=getMisfit(dx);

    % step vector  - eq.(17a)
    sv1=dx-dL;
    sv=[sv1 sv];
    ssv=size(sv);
    svN=[getMisfit(sv1,0) svN];
    if ssv(2)> m+1,  % only m+1 last 'sv' is remembered
        sv(:,m+2)=[];
        svN(m+2)=[];
    end
    %linearity vector - eq.(17b)
    ciXX=repmat(ci',nDat,1);
    lv=dx-dL-sum(ciXX.*ek,2);

    r1=  sum(lv.*dL)/(sqrt(sum(lv.*lv))*sqrt(sum(dL.*dL)));
    r=max(r1,0.5);




    % random re-initialization, but the best
    % eq. (18)
    %%%    if ssv(2) >= m+1 && all(getMisfit(sv,0) < max(nDat-m,1.0))
    if ssv(2) >= m+1
        c=polyfit(1:m+1,svN,1);
    end
    if ssv(2) >= m+1 && c(1) <= 0  % if trend of svN is negative, then reinitialise Simplex
        [par,dat]=make_init(m,kRed,ParDef);
        par(:,m+1)=parBest;
        dat(:,m+1)=datBest;
        sv=[];
        svN=[];
        run_prt('msg',['Reinitialisation of the Simplex. (kRed=',num2str(kRed),')']);
        newSimplex=1;

    elseif Mx < Mdat(1)  % the best is replaced
        par=[px par];
        par(:,m+2)=[];
        dat=[dx dat];
        dat(:,m+2)=[];

        if Mbest > Mx
            datBest=dx;
            parBest=px;
            Mbest=Mx;
            str=sprintf('Improved old: iLoop=% 4d,  iFwd=% 5d,     Mbest=%e ',iLoop,nFwd,Mbest);
            disp(str);
            ccN  =[ccN nFwd];
            ccM  =[ccM Mbest];
            ccPar=[ccPar parBest];
            ccK=[ccK kRed];
        end

    elseif Mx < Mdat(m+1) % the worst is replaced
        par(:,m+1)=px;
        dat(:,m+1)=dx;

    else   % random search of one new model
        Mtmp=Inf;

        for i=1:iTry
            pn=make_init(1,kRed,ParDef);
            dn=hFwd(px);
            nFwd=nFwd+1;
            Mn=getMisfit(dn,0);
            if Mn < Mtmp , pTmp=pn; dTmp=dn; Mtmp=Mn;end
            if Mn <  2*Mdat(1)  % eq.(20)
                kRed=max(kRed-1,0);
                str=sprintf('     kRed=%2d: iLoop=% 4d,  iFwd=% 5d,     Mbest=%e ',kRed,iLoop,nFwd,Mbest);
                disp(str);
                break
            else
                kRed=kRed+1;
                str=sprintf('     kRed=%2d: iLoop=% 4d,  iFwd=% 5d,     Mbest=%e ',kRed,iLoop,nFwd,Mbest);
                disp(str);
            end
        end
        if i >= iTry,
            pn=pTmp;
            dn=dTmp;
            run_prt('wrn','Can`t fullfill condition for new random model.');
            str=sprintf('     kRed=%2d: iLoop=% 4d,  iFwd=% 5d,     Mbest=%e ',kRed,iLoop,nFwd,Mbest);
            disp(str);
        end

        par(:,m+1)=pn;
        dat(:,m+1)=dn;
    end



    % stop condition
    if Mbest < cStop*sqrt(nDat-m) ,
        str='Meet stop condition.';
        run_prt('msg',str);
        disp(str);
        break;
    end

end  % end of main inversion loop

if all(ccN < nFwd)
    ccN  =[ccN nFwd];
    ccM  =[ccM Mbest];
    ccPar=[ccPar parBest];
    ccK  =[ccK kRed];
end

if nargout > 0
    varargout{1}=parBest;
    if nargout == 2
        varargout{2}=[ccN;ccM];
    end
end

%
% final output to protocol file if necessary
%
if strcmp(endMsg,'O.K.'),
    run_prt('ok',endMsg);
end

% plot course of the inversion
figure
subplot(3,1,1),plot(ccN,ccK);
xlabel('nFwd ');ylabel('kRed');
title(['file name: ',Fname]);
subplot(3,1,2),plot(ccN,ccPar);
xlabel('nFwd ');ylabel('parameters');
subplot(3,1,3),plot(ccN,ccM,'k');
xlabel('nFwd ');ylabel('difference');

return
%--------------------------------
function [nDat,Dat]=read_data(in_name)
% read data

global prt_out

if isempty(in_name)
    run_prt('err','No data file  !!!');
end

inData=[in_name,'.dat'];
in=fopen(inData,'r');
if in == -1
    run_prt('err',['No data file >>',in_name,'<< !!!']);
end

line=fgets(in);
n=sscanf(line,'%d');
Dat=fscanf(in,'%e',[1,Inf]);

fclose(in);

if  n < length(Dat)
    str='Given number of data less than actual data size. The data will be cut!!';
    disp(str);
    run_prt('wrn',str);
    Dat(n+1:length(Dat))=[];


elseif n > length(Dat)
    run_prt('err','Given number of data greather than actual data size.');
end

Dat=Dat';
nDat=n;
fprintf(prt_out,'Data file: %s \n',inData);
fprintf(prt_out,'number of data: %d \n\n',nDat);

return
%--------------------------------------------------
function [nPar,ParDef]=read_param(in_name)
% read parameters

global prt_out

inPar=[in_name,'.par'];
if isempty(inPar)
    run_prt('err','No Parameters file.');
end

in=fopen(inPar,'r');
if in == -1
    run_prt('err',['No parameters file >>',inPar,'<< !! !']);
end

line=fgets(in);
n=sscanf(line,'%d');
line=fgets(in);

%default values of Paremeters
ParDef.min(1:n,1)= -1;
ParDef.max(1:n,1)=  1;
ParDef.start(1:n,1)=NaN;

nPar=n;

if line ~= -1
    for i=1:n
        line=fgetl(in);
        if line== -1, break,end
        pos=strfind(line,':');
        if isempty(pos), pos=0;end
        line1=line(pos+1:length(line));
        p=sscanf(line1,'%f');
        if length(p) == 3
            ParDef.min(i)=p(1);
            ParDef.max(i)=p(2);
            ParDef.start(i)=p(3);
        elseif length(p) == 2
            ParDef.min(i)=p(1);
            ParDef.max(i)=p(2);
        elseif length(p) == 1
            ParDef.start(i)=p(1);
        elseif length(p) == 0

        else
            run_prt('err',['Incorrect number of parameters` definition in line ',num2str(i),' !!']);
        end
    end
else
    i=1;
end
fclose(in);

if i==1
    run_prt('wrn','Default values of parameters are used.');
elseif  nPar > i+2
    run_prt('err','Given number of parameters greather than actual parameters size.');
end

fprintf(prt_out,'Parameters file: %s \n',inPar);
fprintf(prt_out,'number of parameters: %d \n\n',nPar);

return
%----------------------------------------------------
function out=open_protocol(name)
% open name.rNN protocol file

t=clock;
cur_dir=pwd;
for i=0:99
    tName=sprintf('%s.r%02d',name,i);
    if ~exist([cur_dir,'\',tName],'file')
        break;
    end
end
out=fopen([cur_dir,'\',tName],'wt');
if out == -1
    error('Error in open porotocol file!!');
end
str=sprintf('Protocol: %s \n',tName);
disp(str);
fprintf(out,'%s \n',str);
fprintf(out,'Start: %s %02d:%02d:%02.3f \n',date,t(4),t(3),t(6));

return
%------------------------------------------------------------
function run_prt(typ,msg)
% write messages (warnings/errors), results and info in protocol file

global prt_out
global nFwd iLoop
global parBest Mbest Mstart
global m

if iLoop > 0
    fprintf(prt_out,'iLoop=% 4d, nFwd=% 5d :',iLoop,nFwd);
end
switch typ
    case 'ok'
        fprintf(prt_out,'END:     %s \n\n',msg);
        str=sprintf('END: iLoop : %d    , nFwd :  %d ',iLoop,nFwd);
        disp(str);
    case 'err'
        fprintf(prt_out,'Error:   %s \n\n',msg);

    case 'wrn'
        fprintf(prt_out,'Warning: %s \n',msg);
        return

    case 'msg'
        fprintf(prt_out,'Message: %s \n',msg);
        return

    case 'wrnEnd'
        fprintf(prt_out,'Warning: %s \n\n',msg);
end


fprintf(prt_out,'So far the best: \n');
fprintf(prt_out,'Mend: %e      (Mstart=%e) \n',Mbest,Mstart);
fprintf(prt_out,'par:');
for i=1:m
    fprintf(prt_out,'%e  ',parBest(i));
end
fprintf(prt_out,' \n\n');

str=sprintf('END: So far the best: ');
disp(str);
str=sprintf('END: Mend: %e      (Mstart=%e) ',Mbest,Mstart);
disp(str);
str=sprintf('END: par:');
for i=1:m
    str=[str sprintf('%f  ',parBest(i))];
end
disp(str);



t=clock;
fprintf(prt_out,'Stop :  %02d:%02d:%02.3f \n',t(4),t(3),t(6));
fprintf(prt_out,'Elapsed time: %s \n\n',num2str(toc));

switch typ
    case 'ok'
        fprintf(prt_out,'O.K. end \n');
        fclose(prt_out);
        disp('End of IM.');
    case 'err'
        fprintf(prt_out,'ERROR end \n');
        fclose(prt_out);
        disp(['ERROR:',msg]);
        error(msg);
    case 'wrnEnd'
        fprintf(prt_out,'WARNING end \n');
        disp(['WARNING:',msg]);
        disp('End of IM.');
end

return
%-----------------------------------------
function [parNew,datNew]=make_init(n,kRed,parDef)
% random initialization of m+1 parameters sets

global hFwd
global m
global parBest
global nFwd


reduc=0.03;   % size parameter space reduction is 3%

% random generation
parNew=rand(m,n);


%scaling & reduction
dP=(parDef.max-parDef.min)/2;
if any(isinf(parBest))
    X0=parDef.min+dP;
else
    X0=parBest;
    dP=dP * power(1.-reduc,kRed);  % the param.space is reduced
end

dPX=repmat(2*dP,1,n);
cP=X0-dP;
cPX=repmat(cP,1,n);
parNew=cPX+parNew.*dPX;

parMinX=repmat(parDef.min,1,n);  % are we in allowed area ?
parMaxX=repmat(parDef.max,1,n);
parNew=max(parNew,parMinX);
parNew=min(parNew,parMaxX);

for i=1:size(parNew,2);
    datNew(:,i)=hFwd(parNew(:,i));
    nFwd=nFwd+1;
end

return
% ----------------------------------------------
function NormMisfit=getMisfit(val,varargin)
% return Norm of misfit


global dat1

etalon=dat1;  % if only 1 input parameter, input data are compared
sD=size(val);
if nargin > 1
    etalon=varargin{1};  % 2nd input parameter is compared
    if etalon == 0
        etalon=zeros(sD(1),1); % 2nd input parameter is 0
    end
end

DatX=repmat(etalon,1,sD(2));
Misfit=DatX-val;

% here norm L2 is used     (can be optionally changed)
NormMisfit=sqrt(sum(Misfit.^2 ));

return
%==============eof===========